#  Copyright 2016 Observable Networks
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals
from datetime import datetime
# from io import StringIO
from json import dumps
from unittest import TestCase

try:
    from unittest.mock import MagicMock, patch
except ImportError:
    from mock import MagicMock, patch

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from kinesis_logs_reader import KinesisLogsReader
from kinesis_logs_reader.__main__ import main as cli_main
from kinesis_logs_reader.utils import gunzip_bytes, gzip_bytes


def _data_message(log_events):
    data = {'messageType': 'DATA_MESSAGE', 'logEvents': []}
    for event in log_events:
        data['logEvents'].append({'extractedFields': event})

    return {'Data': gzip_bytes(dumps(data).encode('utf-8'))}


def _control_message():
    data = {'messageType': 'CONTROL_MESSAGE'}
    return {'Data': gzip_bytes(dumps(data).encode('utf-8'))}


def _create_event(index):
    return {
        'srcaddr': '192.0.2.1',
        'srcport': index,
        'dstaddr': '198.51.100.1',
        'dstport': 443,
        'protocol': 6,
    }


def _get_shard_iterator(**kwargs):
    return {'ShardIterator': '{}_iterator-0001'.format(kwargs['ShardId'])}


DESCRIBE_STREAM = {
    'StreamDescription': {
        'Shards': [{'ShardId': 'shard-0001'}, {'ShardId': 'shard-0002'}],
    }
}


GET_RECORDS = {
    'shard-0001_iterator-0001': {
        'Records': [
            _control_message(),
            _data_message([_create_event(0), _create_event(1)]),
        ],
        'NextShardIterator': 'shard-0001_iterator-0002',
        'MillisBehindLatest': 100,
    },
    'shard-0001_iterator-0002': {
        'Records': [_data_message([_create_event(2), _create_event(3)])],
        'NextShardIterator': 'shard-0001_iterator-0003',
        'MillisBehindLatest': 0,
    },
    'shard-0002_iterator-0001': {
        'Records': [_data_message([_create_event(4), _create_event(5)])],
        'NextShardIterator': 'shard-0002_iterator-0002',
        'MillisBehindLatest': 0,
    },
    'shard-0002_iterator-0002': {
        'Records': [],
        'NextShardIterator': 'shard-0002_iterator-0003',
        'MillisBehindLatest': 0,
    },
}


def _get_client(*args, **kwargs):
    # Mock the boto3 client so tests don't hit the AWS API
    mock_client = MagicMock()
    mock_client.get_paginator.return_value.paginate.return_value = (
        [DESCRIBE_STREAM]
    )
    mock_client.get_shard_iterator.side_effect = _get_shard_iterator
    mock_client.get_records.side_effect = (
        lambda **kwargs: GET_RECORDS[kwargs['ShardIterator']]
    )

    return mock_client


class UtilsTestCase(TestCase):
    def setUp(self):
        self.data = b'Test data'
        self.gz_data = (
            b'\x1f\x8b\x08\x00M\x986W\x02\xff\x0bI-.QHI,I\x04\x00\x11,\xf9Q\t'
            b'\x00\x00\x00'
        )

    def test_gunzip_bytes(self):
        self.assertEqual(gunzip_bytes(self.gz_data), self.data)

    def test_gzip_bytes(self):
        gz_data = gzip_bytes(self.data)
        self.assertEqual(gunzip_bytes(gz_data), self.data)


class KinesisLogsReaderTestCase(TestCase):
    def __init__(self, *args, **kwargs):
        # Python 2 compatibility for tests
        if not hasattr(self, 'assertCountEqual'):
            self.assertCountEqual = self.assertItemsEqual

        return super(KinesisLogsReaderTestCase, self).__init__(*args, **kwargs)

    def setUp(self):
        self.stream_name = 'test-stream'
        self.start_time = datetime(2016, 5, 13, 22, 55, 0)

        self.reader = KinesisLogsReader(
            self.stream_name, kinesis_client=_get_client()
        )

    def test_init(self):
        kwargs = {'region_name': 'test-region', 'profile_name': 'test_profile'}
        patch_path = 'kinesis_logs_reader.kinesis_logs_reader.Session'
        with patch(patch_path) as mock_Session:
            KinesisLogsReader(self.stream_name, **kwargs)
            mock_Session.assert_called_once_with(**kwargs)
            mock_Session.return_value.client.assert_called_once_with('kinesis')

    def test_get_shard_ids(self):
        actual = list(self.reader._get_shard_ids())
        expected = ['shard-0001', 'shard-0002']
        self.assertCountEqual(actual, expected)

    def test_get_shard_iterators(self):
        shard_id = 'shard-0001'
        actual = self.reader._get_shard_iterator(shard_id)
        self.assertEqual(actual, 'shard-0001_iterator-0001')
        self.reader.kinesis_client.get_shard_iterator.assert_called_with(
            StreamName=self.stream_name,
            ShardId=shard_id,
            ShardIteratorType='LATEST',
        )

        shard_id = 'shard-0002'
        actual = self.reader._get_shard_iterator(shard_id, self.start_time)
        self.assertEqual(actual, 'shard-0002_iterator-0001')
        self.reader.kinesis_client.get_shard_iterator.assert_called_with(
            StreamName=self.stream_name,
            ShardId=shard_id,
            ShardIteratorType='AT_TIMESTAMP',
            Timestamp=self.start_time,
        )

    def test_read_shard(self):
        shard_id = 'shard-0001'
        self.reader.shard_iterators[shard_id] = 'shard-0001_iterator-0001'

        # Data messages should have been extracted and returned
        actual_results = list(self.reader._read_shard(shard_id))
        expected_results = [_create_event(0), _create_event(1)]
        self.assertEqual(actual_results, expected_results)

        # The iterator should have advanced
        actual_iterator = self.reader.shard_iterators[shard_id]
        expected_iterator = 'shard-0001_iterator-0002'
        self.assertEqual(actual_iterator, expected_iterator)

        # The response showed the stream as "behind"
        self.assertFalse(self.reader.shard_finished[shard_id])

    def test_reader(self):
        actual = list(self.reader)
        expected = [_create_event(x) for x in range(6)]
        self.assertCountEqual(actual, expected)


@patch(
    'kinesis_logs_reader.kinesis_logs_reader.KinesisLogsReader._get_client',
    _get_client
)
class MainTestCase(TestCase):
    def test_cli_main(self):
        argv = [
            '--start-time="2016-05-15 01:02"',
            '--time-format="%Y-%m-%d %H:%M"',
            'test_stream',
        ]
        with patch('sys.stdout', new_callable=StringIO) as mock_stdout:
            cli_main(argv)
            output = mock_stdout.getvalue().splitlines()

        self.assertEqual(len(output), 7)

        event = _create_event(0)
        event_keys = sorted(event.keys())
        event_values = [event[k] for k in event_keys]

        actual_header = output[0]
        expected_header = '\t'.join(event_keys)
        self.assertEqual(actual_header, expected_header)

        actual_row = output[1]
        expected_row = '\t'.join(str(x) for x in event_values)
        self.assertEqual(actual_row, expected_row)

    def test_cli_count(self):
        argv = ['--count=2', 'test_stream']
        with patch('sys.stdout', new_callable=StringIO) as mock_stdout:
            cli_main(argv)
            output = mock_stdout.getvalue().splitlines()

        self.assertEqual(len(output), 3)



#  Copyright 2016 Observable Networks
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from io import BytesIO
from gzip import GzipFile


def gunzip_bytes(gz_bytes, *args, **kwargs):
    with BytesIO(gz_bytes) as fd:
        with GzipFile(fileobj=fd, mode='rb') as gz_fd:
            return gz_fd.read()


def gzip_bytes(raw_bytes, *args, **kwargs):
    with BytesIO() as fd:
        with GzipFile(fileobj=fd, mode='wb') as gz_fd:
            gz_fd.write(raw_bytes)

        return fd.getvalue()



from __future__ import print_function, unicode_literals
from argparse import ArgumentParser
from datetime import datetime
from itertools import chain
import sys

from .kinesis_logs_reader import KinesisLogsReader


def print_stream(stream_name, start_time, stop_after):
    reader = KinesisLogsReader(stream_name, start_time=start_time)

    # Peek into the first row to determine the header
    first_row = next(reader)
    keys = sorted(first_row.keys())
    print(*keys, sep='\t')

    # Join the first row with the rest of the rows and print them
    iterable = chain([first_row], reader)
    for i, fields in enumerate(iterable, 1):
        print(*[fields[k] for k in keys], sep='\t')
        if i == stop_after:
            break


def main(argv=None):
    argv = argv or sys.argv[1:]

    argument_parser = ArgumentParser('Read logs from AWS Kinesis')
    argument_parser.add_argument(
        'stream_name', type=str, help='Name of the Kinesis stream to read'
    )
    argument_parser.add_argument(
        '--start-time',
        type=str,
        help='Time from which to start reading (default: beginning of stream)'
    )
    argument_parser.add_argument(
        '--count',
        type=int,
        default=0,
        help='Number of records to return (default: 0, which is "no limit")'
    )
    argument_parser.add_argument(
        '--time-format',
        type=str,
        default='%Y-%m-%d %H:%M:%S',
        help='Format string for the --start-time argument',
    )
    args = argument_parser.parse_args(argv)

    if args.start_time:
        start_time = datetime.strptime(args.start_time, args.time_format)
    else:
        start_time = None

    print_stream(args.stream_name, start_time, args.count)


if __name__ == '__main__':
    main()



#  Copyright 2016 Observable Networks
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals
from setuptools import setup, find_packages
import sys

PY2 = sys.version_info[0] == 2


setup(
    name='kinesis_logs_reader',
    version='0.1.0',
    license='Apache',
    url='https://github.com/obsrvbl/kinesis-logs-reader',

    description='Reader for AWS CloudWatch Logs stored in AWS Kinesis',
    long_description=(
        'This project provides a convenient interface for accessing '
        'CloudWatch Logs data stored in AWS Kinesis.'
    ),

    author='Observable Networks',
    author_email='support@observable.net',

    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    entry_points={
        'console_scripts': [
            'kinesis_logs_reader = kinesis_logs_reader.__main__:main',
        ],
    },

    packages=find_packages(exclude=[]),
    test_suite='tests',

    install_requires=['botocore>=1.4.19', 'boto3>=1.3.1'],
    tests_require=['mock'] if PY2 else [],
)



#  Copyright 2016 Observable Networks
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals
from json import loads

from boto3.session import Session

from .utils import gunzip_bytes


class KinesisLogsReader(object):
    def __init__(
        self, stream_name, start_time=None, kinesis_client=None, **kwargs
    ):
        """
        Return an iterable object that will yield log events as Python dicts.
        * `stream_name` identifies the Kinesis stream with a log subscription.
        * `start_time` is a Python datetime object; Records from this Timestamp
        on will be returned. By default the latest records will be returned.
        * `kinesis_client` is a boto3.client object. By default one will be
        created with the given `kwargs`.
        """
        self.kinesis_client = kinesis_client or self._get_client(**kwargs)

        self.stream_name = stream_name
        self.shard_ids = list(self._get_shard_ids())

        self.shard_iterators = {}
        for shard_id in self.shard_ids:
            self.shard_iterators[shard_id] = self._get_shard_iterator(
                shard_id, start_time
            )

        self.shard_finished = {shard_id: False for shard_id in self.shard_ids}
        self.iterator = self._reader()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)

    def next(self):
        # For Python 2 compatibility
        return self.__next__()

    def _get_client(self, **kwargs):
        return Session(**kwargs).client('kinesis')

    def _get_shard_ids(self):
        paginator = self.kinesis_client.get_paginator('describe_stream')
        for page in paginator.paginate(StreamName=self.stream_name):
            stream_description = page.get('StreamDescription', {})
            for shard in stream_description.get('Shards', []):
                yield shard['ShardId']

    def _get_shard_iterator(self, shard_id, start_time=None):
        kwargs = {'StreamName': self.stream_name, 'ShardId': shard_id}
        if start_time is None:
            kwargs['ShardIteratorType'] = 'LATEST'
        else:
            kwargs['ShardIteratorType'] = 'AT_TIMESTAMP'
            kwargs['Timestamp'] = start_time

        response = self.kinesis_client.get_shard_iterator(**kwargs)
        return response['ShardIterator']

    def _read_shard(self, shard_id):
        iterator = self.shard_iterators[shard_id]
        response = self.kinesis_client.get_records(ShardIterator=iterator)

        self.shard_iterators[shard_id] = response['NextShardIterator']
        self.shard_finished[shard_id] = response['MillisBehindLatest'] == 0

        for record in response.get('Records', []):
            gz_data = record['Data']
            raw_data = gunzip_bytes(gz_data)
            data = loads(raw_data.decode('utf-8'))

            if data.get('messageType') != 'DATA_MESSAGE':
                continue

            for flow_record in data.get('logEvents', []):
                yield flow_record['extractedFields']

    def _reader(self):
        while True:
            for shard_id in self.shard_ids:
                for item in self._read_shard(shard_id):
                    yield item

            if all(self.shard_finished.values()):
                break
